#include "layer.hpp"

#include "output.hpp"
#include "popup.hpp"
#include "server.hpp"
#include "surface.hpp"
#include "subsurface.hpp"
#include "types.hpp"

#include "wlr-wrap-start.hpp"
#include <wlr/types/wlr_fractional_scale_v1.h>
#include <wlr/types/wlr_layer_shell_v1.h>
#include <wlr/types/wlr_output_layout.h>
#include <wlr/types/wlr_scene.h>
#include <wlr/types/wlr_subcompositor.h>
#include <wlr/types/wlr_xdg_shell.h>
#include <wlr/util/log.h>
#include "wlr-wrap-end.hpp"

static magpie_scene_layer_t magpie_layer_from_wlr_layer(const zwlr_layer_shell_v1_layer layer) {
	switch (layer) {
		case ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND:
			return MAGPIE_SCENE_LAYER_BACKGROUND;
		case ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM:
			return MAGPIE_SCENE_LAYER_BOTTOM;
		case ZWLR_LAYER_SHELL_V1_LAYER_TOP:
			return MAGPIE_SCENE_LAYER_TOP;
		default:
			return MAGPIE_SCENE_LAYER_OVERLAY;
	}
}

/* Called when the surface is mapped, or ready to display on-screen. */
static void wlr_layer_surface_v1_map_notify(wl_listener* listener, [[maybe_unused]] void* data) {
	wlr_log(WLR_DEBUG, "wlr_layer_surface_v1.events.map(listener=%p, data=%p)", (void*) listener, data);

	Layer& layer = magpie_container_of(listener, layer, map);

	wlr_scene_node_set_enabled(&layer.scene_tree->node, true);
	wlr_surface_send_enter(layer.get_wlr_surface(), &layer.output.wlr);
}

/* Called when the surface is unmapped, and should no longer be shown. */
static void wlr_layer_surface_v1_unmap_notify(wl_listener* listener, [[maybe_unused]] void* data) {
	wlr_log(WLR_DEBUG, "wlr_layer_surface_v1.events.unmap(listener=%p, data=%p)", (void*) listener, data);

	Layer& layer = magpie_container_of(listener, layer, unmap);

	wlr_scene_node_set_enabled(&layer.scene_tree->node, false);
}

/* Called when the surface is destroyed and should never be shown again. */
static void wlr_layer_surface_v1_destroy_notify(wl_listener* listener, [[maybe_unused]] void* data) {
	wlr_log(WLR_DEBUG, "wlr_layer_surface_v1.events.destroy(listener=%p, data=%p)", (void*) listener, data);

	Layer& layer = magpie_container_of(listener, layer, destroy);

	layer.output.layers.erase(std::dynamic_pointer_cast<Layer>(layer.shared_from_this()));
}

static void wlr_layer_surface_v1_commit_notify(wl_listener* listener, [[maybe_unused]] void* data) {
	wlr_log(WLR_DEBUG, "wlr_layer_surface_v1.events.commit(listener=%p, data=%p)", (void*) listener, data);

	Layer& layer = magpie_container_of(listener, layer, commit);

	Server& server = layer.output.server;
	const wlr_layer_surface_v1& surface = layer.wlr;

	const uint32_t committed = surface.current.committed;
	if ((committed & WLR_LAYER_SURFACE_V1_STATE_LAYER) != 0) {
		layer.scene_layer = magpie_layer_from_wlr_layer(surface.current.layer);
		wlr_scene_node_reparent(&layer.scene_tree->node, server.scene_layers[layer.scene_layer]);
	}

	if ((committed & WLR_LAYER_SURFACE_V1_STATE_KEYBOARD_INTERACTIVITY) != 0) {
		switch (layer.wlr.current.keyboard_interactive) {
			case ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_NONE:
				server.try_focus_next_exclusive_layer();
				break;
			case ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND:
			case ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE:
				server.focus_layer(std::dynamic_pointer_cast<Layer>(layer.shared_from_this()));
				break;
		}
	}

	if (committed != 0) {
		layer.output.update_layout();
	}
}

static void wlr_layer_surface_v1_new_popup_notify(wl_listener* listener, void* data) {
	wlr_log(WLR_DEBUG, "wlr_layer_surface_v1.events.new_popup(listener=%p, data=%p)", (void*) listener, data);

	if (data == nullptr) {
		wlr_log(WLR_ERROR, "No data passed to wlr_layer_surface_v1.events.new_popup");
		return;
	}

	Layer& layer = magpie_container_of(listener, layer, new_popup);
	auto* surface = static_cast<Surface*>(layer.wlr.surface->data);
	surface->popups.emplace(std::make_shared<Popup>(*surface, *static_cast<wlr_xdg_popup*>(data)));
}

static void wlr_layer_surface_v1_new_subsurface_notify(wl_listener* listener, void* data) {
	wlr_log(WLR_DEBUG, "wlr_layer_surface_v1.events.new_subsurface(listener=%p, data=%p)", (void*) listener, data);

	if (data == nullptr) {
		wlr_log(WLR_ERROR, "No data passed to wlr_layer_surface_v1.events.new_subsurface");
		return;
	}

	Layer& layer = magpie_container_of(listener, layer, new_subsurface);
	auto& subsurface = *static_cast<wlr_subsurface*>(data);

	layer.subsurfaces.emplace(std::make_shared<Subsurface>(layer, subsurface));
}

Layer::Layer(Output& output, wlr_layer_surface_v1& surface) noexcept
	: listeners(*this), server(output.server), output(output), wlr(surface) {
	const magpie_scene_layer_t chosen_layer = magpie_layer_from_wlr_layer(surface.current.layer);

	scene_tree = wlr_scene_tree_create(output.server.scene_layers[chosen_layer]);
	scene_surface = wlr_scene_layer_surface_v1_create(scene_tree, &surface);
	surface_node = &scene_surface->tree->node;

	surface_node->data = this;
	surface.surface->data = this;

	wlr_fractional_scale_v1_notify_scale(surface.surface, surface.output->scale);
	wlr_surface_set_preferred_buffer_scale(surface.surface, std::ceil(surface.output->scale));

	listeners.map.notify = wlr_layer_surface_v1_map_notify;
	wl_signal_add(&surface.surface->events.map, &listeners.map);
	listeners.unmap.notify = wlr_layer_surface_v1_unmap_notify;
	wl_signal_add(&surface.surface->events.unmap, &listeners.unmap);
	listeners.destroy.notify = wlr_layer_surface_v1_destroy_notify;
	wl_signal_add(&surface.events.destroy, &listeners.destroy);
	listeners.commit.notify = wlr_layer_surface_v1_commit_notify;
	wl_signal_add(&surface.surface->events.commit, &listeners.commit);
	listeners.new_popup.notify = wlr_layer_surface_v1_new_popup_notify;
	wl_signal_add(&surface.events.new_popup, &listeners.new_popup);
	listeners.new_subsurface.notify = wlr_layer_surface_v1_new_subsurface_notify;
	wl_signal_add(&surface.surface->events.new_subsurface, &listeners.new_subsurface);
}

Layer::~Layer() noexcept {
	wl_list_remove(&listeners.map.link);
	wl_list_remove(&listeners.unmap.link);
	wl_list_remove(&listeners.destroy.link);
	wl_list_remove(&listeners.commit.link);
	wl_list_remove(&listeners.new_popup.link);
	wl_list_remove(&listeners.new_subsurface.link);
}

wlr_surface* Layer::get_wlr_surface() const {
	return wlr.surface;
}

Server& Layer::get_server() const {
	return server;
}

bool Layer::is_view() const {
	return false;
}
