#include "xwayland.hpp"

#include "input/seat.hpp"
#include "server.hpp"
#include "types.hpp"
#include "surface/view.hpp"

#include "wlr-wrap-start.hpp"
#include <wlr/util/log.h>
#include "wlr-wrap-end.hpp"

static void ready_notify(wl_listener* listener, [[maybe_unused]] void* data) {
	wlr_log(WLR_DEBUG, "wlr_xwayland.events.ready(listener=%p, data=%p)", (void*) listener, data);

	const std::array<std::string, ATOM_LAST> atom_map = {
		"_NET_WM_WINDOW_TYPE_NORMAL",
		"_NET_WM_WINDOW_TYPE_DIALOG",
		"_NET_WM_WINDOW_TYPE_UTILITY",
		"_NET_WM_WINDOW_TYPE_TOOLBAR",
		"_NET_WM_WINDOW_TYPE_SPLASH",
		"_NET_WM_WINDOW_TYPE_MENU",
		"_NET_WM_WINDOW_TYPE_DROPDOWN_MENU",
		"_NET_WM_WINDOW_TYPE_POPUP_MENU",
		"_NET_WM_WINDOW_TYPE_TOOLTIP",
		"_NET_WM_WINDOW_TYPE_NOTIFICATION",
		"_NET_WM_STATE_MODAL",
	};

	XWayland& xwayland = magpie_container_of(listener, xwayland, ready);

	wlr_xwayland_set_seat(xwayland.wlr, xwayland.server.seat->wlr);

	xcb_connection_t* xcb_conn = xcb_connect(nullptr, nullptr);
	if (const int32_t err = xcb_connection_has_error(xcb_conn)) {
		wlr_log(WLR_ERROR, "XCB connect failed: %d", err);
		return;
	}

	std::array<xcb_intern_atom_cookie_t, ATOM_LAST> cookies = {};
	for (size_t i = 0; i < ATOM_LAST; i++) {
		cookies[i] = xcb_intern_atom(xcb_conn, 0, atom_map[i].length(), atom_map[i].c_str());
	}
	for (size_t i = 0; i < ATOM_LAST; i++) {
		xcb_generic_error_t* error = nullptr;
		xcb_intern_atom_reply_t* reply = xcb_intern_atom_reply(xcb_conn, cookies[i], &error);
		if (reply != nullptr && error == nullptr) {
			xwayland.atoms[i] = reply->atom;
		}
		free(reply);

		if (error != nullptr) {
			wlr_log(WLR_ERROR, "could not resolve atom %s: X11 error code %d", atom_map[i].c_str(), error->error_code);
			free(error);
			break;
		}
	}

	xcb_disconnect(xcb_conn);
}

static void new_surface_notify(wl_listener* listener, void* data) {
	wlr_log(WLR_DEBUG, "wlr_xwayland.events.new_surface(listener=%p, data=%p)", (void*) listener, data);

	if (data == nullptr) {
		wlr_log(WLR_ERROR, "No data passed to wlr_xwayland.events.new_surface");
		return;
	}

	XWayland& xwayland = magpie_container_of(listener, xwayland, new_surface);
	auto& xwayland_surface = *static_cast<wlr_xwayland_surface*>(data);

	xwayland.unmapped_views.emplace_back(std::make_shared<XWaylandView>(xwayland.server, xwayland_surface));
}

XWayland::XWayland(Server& server) noexcept : listeners(*this), server(server) {
	wlr = wlr_xwayland_create(server.display, server.compositor, true);

	listeners.ready.notify = ready_notify;
	wl_signal_add(&wlr->events.ready, &listeners.ready);
	listeners.new_surface.notify = new_surface_notify;
	wl_signal_add(&wlr->events.new_surface, &listeners.new_surface);

	setenv("DISPLAY", wlr->display_name, 1);
}
